@@ -1,6 +1,9 @@
 import contextlib
 import os
 import sys
+import time
+import traceback
+from enum import Enum
 from pathlib import Path
 from typing import Any, AsyncGenerator
 
@@ -8,50 +11,37 @@ import orjson
 import structlog
 from starlette.applications import Starlette
 from starlette.requests import Request
-from starlette.responses import JSONResponse
+from starlette.responses import Response
 from starlette.routing import Route
 from structlog.stdlib import BoundLogger
 
 from ..context import Context, Org, User
-from ..data_api import DataAPI
+from ..data_api import DataAPI, _create_session  # pyright: ignore [reportPrivateUsage]
 from ..invocation_event import InvocationEvent
 from .cloud_event import CloudEventError, SalesforceFunctionsCloudEvent
 from .config import ConfigError, load_config
-from .function_loader import LoadFunctionError, load_function
+from .function_loader import Function, LoadFunctionError, load_function
 from .logging import configure_logging, get_logger
 
 PROJECT_PATH_ENV_VAR = "FUNCTION_PROJECT_PATH"
 
 
-class OrjsonResponse(JSONResponse):
-    """
-    Wrapper around Starlette's `JSONResponse` to use an alternative library for JSON serialisation.
-
-    Used since `orjson` has much better performance than the Python stdlib's `json` module:
-    https://github.com/ijl/orjson#performance
-    """
-
-    def render(self, content: Any) -> bytes:
-        return orjson.dumps(content)
-
-
-async def invoke(request: Request) -> OrjsonResponse:
+async def _handle_function_invocation(request: Request) -> Response:
     """Handle an incoming function invocation request."""
     structlog.contextvars.clear_contextvars()
     logger: BoundLogger = request.app.state.logger
 
     if request.headers.get("x-health-check", "").lower() == "true":
-        return OrjsonResponse("OK")
+        return _make_response("OK", _StatusCode.SUCCESS)
 
     body = await request.body()
 
     try:
         cloudevent = SalesforceFunctionsCloudEvent.from_http(request.headers, body)
     except CloudEventError as e:
-        # TODO: Should the invocation ID be extracted manually for this error message?
         message = f"Could not parse CloudEvent: {e}"
         logger.error(message)
-        return OrjsonResponse(message, status_code=400)
+        return _make_response(message, _StatusCode.REQUEST_ERROR, exception=e)
 
     structlog.contextvars.bind_contextvars(invocationId=cloudevent.id)
 
@@ -69,9 +59,9 @@ async def invoke(request: Request) -> OrjsonResponse:
             base_url=cloudevent.sf_context.user_context.salesforce_base_url,
             domain_url=cloudevent.sf_context.user_context.org_domain_url,
             data_api=DataAPI(
-                cloudevent.sf_context.user_context.org_domain_url,
-                app.state.salesforce_api_version,
-                cloudevent.sf_function_context.access_token,
+                org_domain_url=cloudevent.sf_context.user_context.org_domain_url,
+                api_version=request.app.state.salesforce_api_version,
+                access_token=cloudevent.sf_function_context.access_token,
                 session=request.app.state.data_api_session,
             ),
             user=User(
@@ -82,33 +72,98 @@ async def invoke(request: Request) -> OrjsonResponse:
         )
     )
 
+    function: Function = request.app.state.function
+    function_start_time_ns = time.perf_counter_ns()
+
     try:
-        function_result = await app.state.function(event, context)
+        function_result = await function(event, context)
     except Exception as e:  # pylint: disable=broad-except
         message = (
             f"Exception occurred whilst executing function: {e.__class__.__name__}: {e}"
         )
         logger.exception(message)
-        return OrjsonResponse(message, status_code=500)
+        return _make_response(
+            message,
+            _StatusCode.FUNCTION_ERROR,
+            cloudevent=cloudevent,
+            function_duration_ns=time.perf_counter_ns() - function_start_time_ns,
+            exception=e,
+        )
+
+    function_duration_ns = time.perf_counter_ns() - function_start_time_ns
 
     try:
-        return OrjsonResponse(function_result)
+        return _make_response(
+            function_result,
+            _StatusCode.SUCCESS,
+            cloudevent=cloudevent,
+            function_duration_ns=function_duration_ns,
+        )
     except orjson.JSONEncodeError as e:
         message = (
             f"Function return value cannot be serialized: {e.__class__.__name__}: {e}"
         )
         logger.error(message)
-        return OrjsonResponse(message, status_code=500)
+        return _make_response(
+            message,
+            _StatusCode.FUNCTION_ERROR,
+            cloudevent=cloudevent,
+            function_duration_ns=function_duration_ns,
+            exception=e,
+        )
 
 
-async def handle_internal_error(request: Request, e: Exception) -> OrjsonResponse:
-    message = f"Internal error: {e.__class__.__name__}: {e}"
-    request.app.state.logger.error(message)
-    return OrjsonResponse(message, status_code=500)
+async def _handle_internal_error(request: Request, exception: Exception) -> Response:
+    logger: BoundLogger = request.app.state.logger
+    message = f"Internal error: {exception.__class__.__name__}: {exception}"
+    logger.exception(message)
+    return _make_response(message, _StatusCode.INTERNAL_ERROR, exception=exception)
+
+
+class _StatusCode(Enum):
+    SUCCESS = 200
+    REQUEST_ERROR = 400
+    FUNCTION_ERROR = 500
+    INTERNAL_ERROR = 503
+
+
+def _make_response(
+    content: Any,
+    status_code: _StatusCode,
+    cloudevent: SalesforceFunctionsCloudEvent | None = None,
+    function_duration_ns: int | None = None,
+    exception: Exception | None = None,
+) -> Response:
+    # Based on the `responseExtraInfo` definition in:
+    # https://github.com/forcedotcom/sf-fx-schema/blob/main/schema.json
+    metadata: dict[str, str | int | bool] = {
+        "requestId": cloudevent.id if cloudevent else "n/a",
+        "source": cloudevent.source if cloudevent else "n/a",
+        "statusCode": status_code.value,
+    }
+
+    if function_duration_ns:
+        metadata["execTimeMs"] = round(function_duration_ns / (1000 * 1000))
+
+    if exception:
+        metadata["stack"] = "".join(traceback.format_exception(exception))
+        metadata["isFunctionError"] = status_code == _StatusCode.FUNCTION_ERROR
+
+    # We're not using Starlette's `JSONResponse`, since it uses the Python stdlib's
+    # `json` module for JSON serialization, whereas `orjson` has better performance:
+    # https://github.com/ijl/orjson#performance
+    return Response(
+        content=orjson.dumps(content),
+        media_type="application/json",
+        status_code=status_code.value,
+        headers={
+            "x-extra-info": orjson.dumps(metadata).decode(),
+        },
+    )
 
 
 @contextlib.asynccontextmanager
-async def lifespan(app: Starlette) -> AsyncGenerator[None, None]:
+async def _lifespan(app: Starlette) -> AsyncGenerator[None, None]:
     """
     Asynchronous context manager for handling app setup/teardown.
 
@@ -116,10 +171,7 @@ async def lifespan(app: Starlette) -> AsyncGenerator[None, None]:
     requests, and anything after will be run when the server shuts down.
     """
     configure_logging()
-    # `get_logger()` returns a proxy that only instantiates the logger on first usage.
-    # Calling `bind()` here ensures that this instantiation doesn't have to occur each
-    # time the function is invoked.
-    app.state.logger = get_logger().bind()
+    app.state.logger = get_logger()
 
     # This env var is set by the CLI, as a way to propagate CLI args to the ASGI app.
     project_path = Path(os.environ[PROJECT_PATH_ENV_VAR])
@@ -136,18 +188,16 @@ async def lifespan(app: Starlette) -> AsyncGenerator[None, None]:
 
     app.state.salesforce_api_version = config.salesforce_api_version
 
-    async with (
-        DataAPI._create_session()  # pyright: ignore [reportPrivateUsage] pylint:disable=protected-access
-    ) as data_api_session:
+    async with _create_session() as data_api_session:
         app.state.data_api_session = data_api_session
         yield
 
 
 # The ASGI app that will be run by uvicorn.
 app = Starlette(
-    exception_handlers={Exception: handle_internal_error},
-    lifespan=lifespan,
+    exception_handlers={Exception: _handle_internal_error},
+    lifespan=_lifespan,
     routes=[
-        Route("/", invoke, methods=["POST"]),
+        Route("/", _handle_function_invocation, methods=["POST"]),
     ],
 )
