@@ -10,23 +10,47 @@ class User:
     """Information about the invoking Salesforce user."""
 
     id: str
-    """The user's ID."""
+    """
+    The user's ID.
+
+    For example: `005JS000000H123`
+    """
     username: str
-    """The name of the user."""
+    """
+    The username of the user.
+
+    For example: `user@example.tld`
+    """
     on_behalf_of_user_id: str | None
-    """The ID of the user on whose behalf this user is operating."""
+    """
+    The ID of the user on whose behalf this user is operating.
+
+    For example: `005JS000000H456`
+    """
 
 
 @dataclass(frozen=True, kw_only=True, slots=True)
 class Org:
-    """Information about the invoking Salesforce organization and user."""
+    """Information about the invoking Salesforce org and user."""
 
     id: str
-    """The Salesforce organization ID."""
+    """
+    The Salesforce org ID.
+
+    For example: `00DJS0000000123ABC`
+    """
     base_url: str
-    """The base URL of the Salesforce organization."""
+    """
+    The base URL of the Salesforce org.
+
+    For example: `https://example-base-url.my.salesforce-sites.com`
+    """
     domain_url: str
-    """The domain URL of the Salesforce organization."""
+    """
+    The domain URL of the Salesforce org.
+
+    For example: `https://example-domain-url.my.salesforce.com`
+    """
     data_api: DataAPI
     """An initialized data API client instance for interacting with data in the org."""
     user: User
@@ -38,4 +62,4 @@ class Context:
     """Information relating to the function and the Salesforce org with which it is associated."""
 
     org: Org
-    """Information about the invoking Salesforce organization and user."""
+    """Information about the invoking Salesforce org and user."""
