@@ -3,10 +3,10 @@ from dataclasses import dataclass
 __all__ = ["ReferenceId"]
 
 
-@dataclass(frozen=True, slots=True)
+@dataclass(frozen=True, kw_only=True, slots=True)
 class ReferenceId:
     """
-    A reference id for an operation inside a unit of work.
+    A reference id for an operation inside a `UnitOfWork`.
 
     Used to reference results of other operations inside the same unit of work.
     """
